//	file: detour.mod
//
// TITLE: Learning to Detour.
// GOAL:  Schema Construction and Dynamics based on SBL.
//
//	REFERENCES: 
//	Refer to Corbacho and Arbib (1995) for details on detour itself.
//	Refer to Corbacho and Arbib (1997) for further details on Learning.
//
//	ARCHITECTURE: 
//	Perceptual Schemas as functional units.
//	Fields corresponding to schema instances;
//	Integration of fields; WTA; Sensorimotor Xform;
//	Motor schemas as functional units.
//
//	SCHEMA DYNAMICS:
//	Relaxation Labeling (Hummel and Zucker, 1983).
//	Control of schemas based on distributed C&C dynamics.
//	Inclusion of NeuroSchemas.
//
//	SIMPLIFICATIONS:
//	Perceptual Schemas (not Neural Implementation).
//	Motor Schemas as functional units.
//
//	SCHEMA CONSTRUCTION:
//	Autonomous Schema Construction in Run time.
// 	Coherence and Performance Maximization.
//
//	AUTHOR: Fernando Jose Corbacho
// DATE (last update): July 3rd, 98        (prev): July 2nd, 98.
//

nslClass DetourLib
{				
static void gauss1D(NslFloat1 kernel, NslFloat0 sigma)
{
  int xn = kernel.get_imax();
  float sig = sigma.getValue();
  float k = xn/2;
  const float PI = 3.14159;
  float f= sig*nslSqrt(2.0*PI);

	for (int i=-k; i <= k - (1 - xn%2); i++)
	    kernel[i+k] = nslExp(-(i*i)/(2*sig*sig))/f;
}

static void gauss2D(NslFloat2 kernel, NslFloat0 sigma)	
{
  int xn = kernel.get_imax();
  int yn = kernel.get_jmax();

  float sig = sigma.get_data();
  int k1 = xn/2;
  int k2 = yn/2;
  const float PI = 3.14159;
  float f= sig*nslSqrt(2.0*PI);

  for (int i=-k1; i<=k1 - (1 - xn%2); i++)
    for (int j=-k2; j<=k2 - (1 - yn%2); j++)
      kernel[i+k1][j+k2] = nslExp(-(i*i+j*j)/(2*sig*sig))/f;
}

static NslFloat1 maxim(NslFloat1 u)
// returns the vector normalized (substraction) by its maximum.
// only the maximum is above threshold (by 0.01).
{
	int imax = u.get_imax();
	NslFloat1 c(imax);

	float max = 0;			// Maximum value
	int max_i=0;			// Coordinates of max value
	
	max = nslMax(u);
	if (max == 0)
	 	for (int i=0; i < imax; i++)
			c[i]= - 0.01;
	else
		for (int i=0; i < imax; i++)
			c[i]= u[i] - max + 0.01;

	c = nslStep(c);
	return c;
}

static NslFloat2 filter(NslFloat2 view, int f)
// filters out all the elements in the matrix that do not have a 
// value of "f". Simplified feature detector where "f" is the feature.
{
	int i,j;
	int imax = view.get_imax();
	NslFloat2 c(imax, imax);

	for (i=0; i<imax; i++)
	  for (j=0; j<imax; j++)
		if (view[i][j] == f)
			c[i][j] = 1;
		else 
			c[i][j] = 0;	

	return c;
}
}

////////////////////////////// ENVIRONMENT /////////////////////////////////

nslModule World (int wx,int wy, int wz) 
{
	public NslDoutFloat2 worldXZ(wx, wy);  
	public NslDinFloat0 wangle();			// view angle
	public NslDinFloat0 yw(), xw(),zw();	// prey coordinates
	public NslDinFloat0 xb_init(), xb_end(), yb(),	zb();	// barrier coordinates
	public NslDinFloat0 xf(), yf(), zf();	// frog coordinates 
	private NslFloat2 worldXY(wx, wz);
	private NslInput3d in(wx,wy,wx,wz);

public void initSys()
{
	worldXY = 0;
	worldXZ = 0;
}

public void simRun()
{
	in.set_data("frog.transx",xf);
	in.set_data("frog.transy",yf);
	in.set_data("frog.xc",xf);
	in.set_data("frog.yc",yf);
	in.set_data("frog.angle",wangle);

	in.run();
	worldXZ = in.get_xzview();
}
}

nslModule Prey ()

{
	public NslDoutFloat0 xw(), yw(), zw();	
	private NslFloat0 speed();		
	private NslFloat0 boundary();

public void simRun()
{
	if (xw < boundary) 
		xw = xw + speed; 	
	speed = speed * -1;		// Oscillatory behavior.
}
}

nslModule VisualField (int wx,int wz,int recsize) 
{
	public NslDoutFloat2 visualField(recsize, recsize);
	public NslDinFloat2 worldXZ(wx,wz);
	public NslDinFloat0 zf(), xf();	

public void simRun()
{
	int recsize = visualField.getSize();
	int isize = worldXZ.getRows();
	int jsize = worldXZ.getCols();
	
	visualField = worldXZ.getSector(isize-recsize,isize-1,
		jsize/2-recsize/2,jsize/2+recsize/2-1);
}
}

nslModule Depth ()
{
	public NslDinFloat0 yf(), xf();
	public NslDinFloat0 yw(), xw();
	public NslDinFloat0 yb(), xb_init(), xb_end();
	public NslDoutFloat0 depthY(),depthX();
	public NslDoutFloat0 ps();
	private NslFloat0 th();	
	private NslFloat0 safeDistance();

public void initSys()
{
	depthY = 0;
	depthX = 0;
	ps = 0;
}

public void simRun()
{
	if (yf > yb)
		depthY = yf - yb;
	else 
		depthY = 0;

	if (xf > xb_end)
		depthX = xf - xb_end;
	else if (xf < xb_init)
		depthX = xb_init - xf;
	else
		depthX = 0;


	if (depthY > safeDistance || depthX > safeDistance)
		ps = 2;			// Go up fast
	else 
		ps = 0.0;
}
}

nslModule Tactile ()
{
	public NslDinFloat0 depthY(), depthX();
	public NslDoutFloat0 ps();
	private NslFloat0 th();	
	private NslFloat0 safeDistance();

public void initSys()
{
	ps = 0;
}

public void simRun()
{
	if (depthY > 0 && depthY <= safeDistance && depthX <= safeDistance)      
		ps = 1;		// Go up fast
	else 
		ps = 0.0;		
}
}

nslModule PreyRec (int recsize,int tx,int ty)
{
	public NslDinFloat2 visualField(recsize, recsize);
	public NslDoutFloat2 t5_2f(tx,ty);
	public NslDoutFloat0 ps();
	private NslFloat2 t5_2(tx,ty);
	private NslFloat0 th();

public void initSys()
{
	ps = 0;
}

public void simRun()
{	
	t5_2 = DetourLib.filter(visualField,2);		
	if (nslSum(t5_2) >= 1) 
		ps=0.9;								  
	else 
		ps=0;

	ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	
	if (ps > th) 
		t5_2f = nslRamp(t5_2);			
	else
		t5_2f = 0;
}

nslModule SorRec (int recsize,int tx,int ty)
{
	public NslDinFloat2 visualField(recsize, recsize);
	public NslDoutFloat2 th10f(tx,ty);
	public NslDinFloat0 depthY(), depthX();	
	public NslDoutFloat0 ps();
	private NslFloat2 th10(tx,ty);				
	private NslFloat0 edge();
	private NslFloat0 th();

public void initSys()
{
	ps = 0;
}

public void simRun()
{	
   th10 = DetourLib.filter(visualField, 1);	 
	if (nslSum(th10) >= 1) 
		ps=0.9;
	else 
		ps=0;  
		
	ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	
	if (ps > th) 
		th10f = nslRamp(th10); 
	else
		th10f = 0;
}
}

nslModule PreyApproach (int tx,int ty)  
{
	public NslDinFloat2 t5_2f(tx,ty);			
	public NslDinFloat0 psfo();
	public NslDoutFloat0 ps();
	public NslDoutFloat1 preyHor_f(tx);		
	private NslFloat1 preyHor(tx);		// Parcellation		(1D).
	private NslFloat1 preyHor_default(tx);
	private NslFloat2 preyF(tx, ty);		// PREY ATRACTANT FIELD (2D).
	private NslFloat0 prey_norm();		// Normalization
	private NslFloat2 t5_2_rf(80, 80);	// KERNEL for Prey Attractant Field.
	private NslFloat2 t5_2_erf(80, 80); 
	private NslFloat0 t5_2_erf_sig();
	private NslFloat0 t5_2_erf_wgt();
	private NslFloat0 rsfo();
	private NslFloat0 th();	

public void initSys()
{
	rsfo = 1.0;	// Prey & Prey Approach.
	ps = 0;

	DetourLib.gauss2D(t5_2_erf,t5_2_erf_sig);
	t5_2_rf = t5_2_erf_wgt * t5_2_erf;
}

public void simRun()
{
	preyF = t5_2_rf * t5_2f;		
	preyHor = nslReduceRow(preyF);	// Parcellation: horizontal comp.

	ps = rsfo*psfo; 
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	

	if (ps > th) {
		float mx = nslMax(preyHor);	
		if (mx != 0.0)
			preyHor_f = preyHor/mx;	// Normalize.
	}
	else
		preyHor_f = 0;
}
}

nslModule SorAvoid (int tectsize,int tx,int ty,int rrf)  
{
	public NslDinFloat2 th10f(tx,ty);
	public NslDoutFloat1 gapsf(tectsize);	
	public NslDinFloat0 psso();
	public NslDoutFloat0 ps();
	private NslFloat1 th10Hor(tx);		// Parcellation		
	private NslFloat1 tm_irf(rrf*2);
	private NslFloat1 tm_rf("rrf*2);
	private NslFloat0 tm_irf_sig();
	private NslFloat0 tm_irf_wgt();
	private NslFloat1 gaps(tectsize);		// openings within obstacles.
	private NslFloat0 rsso();		
	private NslFloat0 th();

public void initSys()
{
	rsso = 1.0;	// Obstacle & Obstacle Avoid
	ps = 0;

	DetourLib.gauss1D(tm_irf,tm_irf_sig);
	tm_rf = - tm_irf_wgt * tm_irf;
}

public void simRun()
{
	th10Hor = nslReduceRow(th10f); // Parcellation (from 2D to 1D)
	gaps = tm_rf * th10Hor;		// Convolve with kernel	
				
	ps = rsso*psso; 
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	

	if (ps > th)	{
		float mx = nslMax(gaps*-1);	
		if (mx != 0.0)
			gapsf = gaps/mx;		
	}
	else 
		gapsf = 0;
}
}

nslModule BumpAvoid (int tectsize) 
{
	public NslDinFloat0 depthY(),depthX();
	public NslDoutFloat1 baf(tectsize);			
	public NslDinFloat1 gaps(tectsize);
	public NslDoutFloat1 gapsf(tectsize);
	public NslDoutFloat0 ps();
	public NslDinFloat0 psta();
	private NslFloat0 field_A();
	private NslFloat0 field_center();
	private NslFloat0 tune_tm();	// TUNE GAPS field (avoid bumping).
	private NslFloat0 tune_tm_base();
	private NslFloat1 tune_tm_layer(tectsize);
	private NslFloat0 safeDistance();
	private NslFloat0 rsta();
	private NslFloat0 th();

public void initSys()
{
	rsta = 1.0;	// Tactile & Bumping Avoid
	ps = 0;
}

public void simRun()
{
	// both should be larger than zero for no bump
	if (depthY <= safeDistance && depthX <= safeDistance) 	//Bumping.   
		ps = 0;
	else
		ps = -1.0;		// Go down fast: -2.0

	if (tune_tm < 1.5)		// saturate tune_tm.
			tune_tm = tune_tm + tune_tm_base;

	tune_tm_layer = tune_tm;
	tune_tm_layer = tune_tm_layer ^ nslStep(-gaps);	// Modulate only already active neurons.

	gaps = gaps - tune_tm_layer; // gaps fields gets tuned.
	gapsf = gaps;

	ps = ps + rsta*psta;
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	

// every bump it increases tune_tm until it reaches a saturation point.
	if (ps > th){
		field_center = field_center +.3;	// Tunes baf by increasing eccentricity.
		baf[(int)field_center.getValue()] = field_A; // Generate Bumping Avoidance FIELD.
	}
	else
		baf = 0;
}
}

nslModule MHM (int tectsize) 
{
	public NslDinFloat1 baf(tectsize);			
	public NslDinFloat1 gapsf(tectsize);
	public NslDinFloat1 preyHor_f(tectsize);			
	public NslDoutFloat1 mhm(tectsize);
	public NslDoutFloat0 ps();
	public NslDinFloat0 pspa();
	public NslDinFloat0 psoa();
	private NslFloat1 mhm_hat(tectsize);		// Predictive MHM.
	private NslFloat0 field_A();
	private NslFloat0 field_center();
	private NslFloat0 rspa();
	private NslFloat0 rsoa();

public void initSys()
{
	rspa = 1.0;	
	rsoa = 1.0;	
	ps = 0;
}
public void simRun()
{
	// Dynamics of Motor Heading Map (MHM): integrating several fields.
	// New fields can also be added while learning.

	if (d_mhm > d_norm)
	{			 
		baf[field_center.getValue()] = field_A;
		in = baf; // New Field "inserted"
	}
	else
		in = 0; // reset input (cf. antidromic
		
	mhm_hat = mhm;					// Predictive MHM.
	mhm = gapsf + preyHorF + baf + in;		// Fields over MHM.

	ps = ps + rspa*pspa + rsoa*psoa;
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	
}
}

nslModule WTA(int tectsize) 
{
	public NslDinFloat1 mhm(tectsize);
	public NslDoutFloat1 wta_mhm(tectsize);

public void simRun()
{		
	wta_mhm = DetourLib.maxim(mhm);	
}
}

nslModule Xform (int tectsize) 
{
	public NslDinFloat1 wta_mhm(tectsize);	// Predictive MHM.
	public NslDoutFloat0 angle();			// Heading (target) angle.
	public NslDoutFloat0 ps();
	public NslDinFloat0 pspa();
	public NslDinFloat0 psoa();
	private NslFloat1 angle_position(tectsize);
	private NslFloat0 rspa();
	private NslFloat0 rsoa();

public void initSys()
{
	rspa = 1.0;	
	rsoa = 1.0;	
	ps = 0;
}
public void simRun()
{
// Sensory motor Transformation from retinotopic coordinates to population coding.

	int i = nslMaxValue(wta_mhm); 
	angle = 0;
	if (i != 0)
		angle = i - wta_mhm.getRows()/2; 

	ps = ps + rspa*pspa + rsoa*psoa;
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	
}

nslModule Forward()
{
	public NslDinFloat0 depthY();				// Heading (target) angle.
	public NslDoutFloat0 out(); 
	public NslDoutFloat0 ps();
	public NslDinFloat0 psfo();
	public NslDinFloat0 psta();
	public NslDinFloat0 psdp();
	public NslDinFloat0 psmhm();
	public NslDinFloat0 psbk();
	private NslFloat0 step();
	private NslFloat0 rsfo();
	private NslFloat0 rsta();
	private NslFloat0 rsdp();
	private NslFloat0 rsmhm();
	private NslFloat0 rsbk();
	private NslFloat0 th();

public void initSys()
{
	rsfo = 0.5;
	rsta = -1.0;
	rsdp = 1.0;			
	rsmhm = 0.5;	
	rsbk = -1.0;
		
	ps = 0;
}
public void simRun()
{			
	ps = ps + rsfo*psfo + rsta*psta + rsdp*psdp + rsmhm*psmhm + rsbk*psbk;
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	

	if (ps > th) {
        ps = -1.0;				// Reset        
        out = step;
	}
	else
		out = 0;
}
}

nslModule Sidestep ()
{
	public NslDinFloat0 angle();				// Heading (target) angle.
	public NslDoutFloat0 out();	
	public NslDinFloat0 psso();
	public NslDinFloat0 psta();
	public NslDinFloat0 psba();
	public NslDinFloat0 psmhm();
	public NslDinFloat0 psx();
	public NslDoutFloat0 ps();
	private NslFloat0 rsso();
	private NslFloat0 rsta();
	private NslFloat0 rsba();
	private NslFloat0 rsmhm();
	private NslFloat0 rsx();
	private NslFloat0 th();	

public void initSys()
{
	rsso = 0.5;
	rsta = 1.0;
//	rsdp = -1.0;	
	rsba = 0.5;			
	rsmhm = 0.5;	
	rsx = 0.5;	
		
	ps = 0;
}
public void simRun()
{
	ps = ps + rsso*psso + rsta*psta + rsba*psba + rsmhm*psmhm + rsx*psx;
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	

	if (ps > th)	 {
         ps = 0; // -1.0;			// Reset
		out = angle;
	}
	else
		out = 0;
}
}

nslModule Backward ()
{
	public NslDoutFloat0 out();
	public NslDoutFloat0 ps();
	public NslDinFloat0 psta();
	public NslDinFloat0 psba();
	public NslDinFloat0 psfw();
	private NslFloat0 step();	
	private NslFloat0 rsta();
	private NslFloat0 rsba();
	private NslFloat0 rsfw();
	private NslFloat0 th();

public void initSys()
{
	rsta = 1.0;
	rsba = 1.0;		
	rsfw = -1.0;	
		
	ps = 0;
}
public void simRun()
{		
	ps = ps + rsta*psta + rsba*psba + rsfw*psfw;
   ps = nslSigma(ps,-1.0,1.0,-1.0,1.0);	

	if (ps > th)	 {
        ps = -1.0;			// Reset
		out = step;
	}
	else
		out = 0;
}	
}

// Frog

nslModule Frog (int wx,int wz,int recsize,int tectsize,int tx,int ty,int rrf) 
{
	public NslDinFloat2 worldXZ(wx,wz);  
	public NslDoutFloat0 xf(), yf(), zf();	
	public NslDinFloat0 xw(), yw(), zw();	
	public NslDinFloat0 xb_init(), xb_end(), yb(),	zb();
	
	public NslDoutFloat0 forward_step();
	public NslDoutFloat0 backward_step();
	public NslDoutFloat0 wangle();
	public NslDoutFloat0 angle();

	private VisualField visual(wx,wz,recsize);
	private Depth depth();
	private Tactile tactile();
	private PreyRec preyRec(recsize,tx,ty);
	private SorRec sorRec(recsize,tx,ty);
	private PreyApproach preyApproach(tx,ty);
	private SorAvoid sorAvoid(tectsize,tx,ty,rrf);
	private BumpAvoid bumpAvoid(tectsize);
	private MHM mhm(tectsize);
	private WTA wta(tectsize);
	private Xform xform(tectsize);
	private Forward forward();
	private Sidestep sidestep();
	private Backward backward();
			
	private Float0 safeDistance();

public void makeConn()
{
	nslRelabel (worldXZ,visual.worldXZ);
	nslRelabel (xb_init,depth.xb_init);
	nslRelabel (xb_end,depth.xb_end);
	nslRelabel (yb,depth.yb);
	nslRelabel (sidestep.out,angle);
	nslRelabel (forward.out,forward_step);
	nslRelabel (backward.out,backward_step);
	nslRelabel (xw,depth.xw);	
	nslRelabel (yw,depth.yw);

	nslConnect (xf,visual.xf);
	nslConnect (zf,visual.zf);
	nslConnect (xf,depth.xf);	
	nslConnect (yf,depth.yf);

	nslConnect (visual.visualField,preyRec.visualField);
	nslConnect (visual.visualField,sorRec.visualField);
	nslConnect (depth.ps,forward.psdp);	
	nslConnect (depth.depthY,tactile.depthY);
	nslConnect (depth.depthX,tactile.depthX);
	nslConnect (depth.depthY,sorRec.depthY);
	nslConnect (depth.depthX,sorRec.depthX);
	nslConnect (depth.depthY,bumpAvoid.depthY);
	nslConnect (depth.depthX,bumpAvoid.depthX);
	nslConnect (depth.depthY,forward.depthY);
	nslConnect (tactile.ps,bumpAvoid.psta);	
	nslConnect (tactile.ps,forward.psta);	
	nslConnect (tactile.ps,sidestep.psta);	
	nslConnect (tactile.ps,backward.psta);	
	nslConnect (preyRec.t5_2f,preyApproach.t5_2f);
	nslConnect (preyRec.ps,preyApproach.psfo);		
	nslConnect (preyRec.ps,forward.psfo);	
	nslConnect (sorRec.th10f,sorAvoid.th10f);
	nslConnect (sorRec.ps,sorAvoid.psso);	
	nslConnect (sorRec.ps,sidestep.psso);
	nslConnect (preyApproach.preyHor_f,mhm.preyHor_f);		
	nslConnect (preyApproach.ps,mhm.pspa);	
	nslConnect (preyApproach.ps,xform.pspa);	
	nslConnect (sorAvoid.gapsf,bumpAvoid.gaps);		
	nslConnect (sorAvoid.ps,mhm.psoa);	
	nslConnect (sorAvoid.ps,xform.psoa);	
	nslConnect (bumpAvoid.baf,mhm.baf);		
	nslConnect (bumpAvoid.ps,sidestep.psba);	
	nslConnect (bumpAvoid.ps,backward.psba);	
	nslConnect (bumpAvoid.gapsf,mhm.gapsf);		
	nslConnect (mhm.ps,forward.psmhm);	
	nslConnect (mhm.ps,sidestep.psmhm);	
	nslConnect (mhm.mhm,wta.mhm);	
	nslConnect (wta.wta_mhm,xform.wta_mhm);	
	nslConnect (xform.angle,sidestep.angle);	
	nslConnect (xform.ps,sidestep.psx);	
	nslConnect (forward.ps,backward.psfw);	
	nslConnect (backward.ps,forward.psbk);	
}

public void endRun()
{
	double pi = 3.1415926535;
	double iangle = 0;

	double x0 = xf.getValue() - xb_init.getValue();
	double x2 = xb_end.getValue() - xf.getValue();
	double yd = yf.getValue() - yb.getValue();

	double x1 = xf.getValue() - xw.getValue();
	double y1 = yf.getValue() - yw.getValue();

	double xd = angle.getValue() - yd*x1/y1; 

	double t = atan(xd/yd);
	double disp = forward_step.getValue() - backward_step.getValue(); // overall displacement
	double dispx = fabs(disp)*sin(t);
	double dispy = disp*cos(t);
	double dispx0 = dispx;
	double dispy0 = dispy;

	if (yf > yb && (yf - dispy < yb + safeDistance) && 
		(xf + dispx > xb_init - safeDistance) && 
		(xf + dispx < xb_end + safeDistance)) 
	{
		dispy = yf.getValue() - yb.getValue() - safeDistance; // maximum movement
		dispx = dispx0*dispy/dispy0;
	}

	yf = yf - dispy;

	if ((yf >= yb - safeDistance) && (yf <= yb + safeDistance))
	{
		if ((xf + dispx < xb_init) && (xf + dispx > xb_init - safeDistance))
			dispx = xb_init.getValue() - xf.getValue() - safeDistance.getValue(); // maximum movement
		else if ((xf + dispx > xb_end) && (xf + dispx < xb_end + safeDistance))
			dispx = xf.getValue() - xb_end.getValue() + safeDistance; // maximum movement
	}

	xf = xf + dispx; 

	if (yf <= yw)
	{
		nslPrintln("Ate Worm");
		yf = yw;
		xf = xw; 

		system.breakEpoch();
		return;
	}

	double x = 0, y = 0;

	x = xf.getValue() - xw.getValue();
	y = yf.getValue() - yw.getValue();

	if (y != 0) {
		iangle = atan(x/y);
		wangle = iangle*180/pi;
	}
	else {
		iangle = pi/2;
		wangle = 90;
	}
}
}

nslModel Detour ()
{ 
	private World world(200,200,200);
	private Prey prey();
 	private Frog frog(200,200,80,80,80,80,40);

public void makeConn()
{
	nslConnect(world.worldXZ,frog.worldXZ);
	nslConnect(world.xb_init,frog.xb_init);
	nslConnect(world.xb_end,frog.xb_end);
	nslConnect(world.yb,frog.yb);
	nslConnect(world.zb,frog.zb);
	
	nslConnect(frog.xf,world.xf);
	nslConnect(frog.yf,world.yf);
	nslConnect(frog.zf,world.zf);
	nslConnect(frog.wangle,world.wangle);

	nslConnect(prey.xw,frog.xw);
	nslConnect(prey.yw,frog.yw);
	nslConnect(prey.zw,frog.zw);

	nslConnect(prey.xw,world.xw);
	nslConnect(prey.yw,world.yw);
	nslConnect(prey.zw,world.zw);
}
}

